import uuid
from copy import deepcopy
from importlib.resources import files
from typing import Any

from .. import RELAIData
from ..data.data_type import DataType
from .feedback_question import FeedbackQuestion


class FeedbackQuestions(RELAIData):
    """Data model for a feedback questions component.

    Args:
        feedback_questions (list[dict]): List of questions to be asked. Each question must be a dictionary in the
            following format:
            {
                "question": str,
                "type": Type of the question. Must be either a `DataType` object or the string "contextSelection".
                "context": A list of RELAIData components to show as context for the question. Required if the type is
                    "contextSelection".
                "modifies": A list of one or more strings that uniquely determine what needs to be changed in the
                    application result. If not provided, a unique ID will be generated for each question.
            }
        feedback (list[Any], optional): Feedback for each of the questions. In most cases, this will be `None`,
            especially when the component is created in a RELAI application, as the user has not provided feedback
            yet. Defaults to None.
        column_span (int, optional): Number of columns the component should span. Defaults to 12.
    """

    schema_file: str = files("relai_visualization").joinpath("schema/components/feedback-questions.schema.json")

    def __init__(
        self,
        feedback_questions: list[dict],
        feedback: list[Any] = None,
        column_span: int = 12,
    ) -> None:
        if feedback and len(feedback) != len(feedback_questions):
            raise ValueError("Number of feedbacks must match the number of feedback questions.")

        if not feedback:
            feedback = [None] * len(feedback_questions)

        feedback_questions_dict = {}
        feedback_dict = {}
        for question, answer in zip(feedback_questions, feedback):
            question_id = f"question-{uuid.uuid4().hex}"
            question_modifies = [question_id]
            try:
                # Add question ID to the list of `modifies` identifiers
                question_modifies += question["modifies"]
            except KeyError:
                pass
            feedback_questions_dict[question_id] = FeedbackQuestion(
                question=question["question"],
                type=question["type"],
                modifies=question_modifies,
                context=question.get("context"),
            ).data
            feedback_dict[question_id] = answer

        super().__init__(
            componentName="feedback-questions",
            columnSpan=column_span,
            feedbackQuestions=feedback_questions_dict,
            feedback=feedback_dict,
        )

    def custom_validation(self) -> None:
        if self.data["feedback"]:
            for question_idx, (question_id, answer) in enumerate(self.data["feedback"].items()):
                if question_id not in self.data["feedbackQuestions"]:
                    raise ValueError(f"Feedback question with ID {question_id} does not exist.")
                if answer and not DataType.from_dict(self.data["feedbackQuestions"][question_id]["type"]).validate(
                    answer
                ):
                    raise ValueError(f"Invalid feedback {answer} for question {question_idx}.")

    @classmethod
    def user_feedback(cls, component: dict) -> dict:
        """Get the user feedback for the feedback questions.

        Args:
            component (dict): Feedback questions component data as a dictionary.

        Returns:
            dict: User feedback for the feedback questions as a dictionary mapping question to feedback values.
        """
        output = {}
        for question_id, answer in component["feedback"].items():
            output[component["feedbackQuestions"][question_id]["question"]] = answer
        return output

    @classmethod
    def update_with_feedback(cls, component: dict, modifies: list[str], new_value: Any) -> dict:
        """Update the feedback questions component with the new feedback value.

        Args:
            component (dict): Feedback questions component to update.
            modifies (list[str]): Identifiers for the feedback question to update. Only the first element of the list
                is used to identify the question within this component. The rest of the elements may be used to
                identify elements in other components.
            new_value (Any): New feedback value.

        Returns:
            dict: Updated feedback questions component.
        """
        updated_component = deepcopy(component)
        # First element is the question ID, the rest are for other components and are ignored here
        question_id = modifies[0]
        updated_component["feedback"][question_id] = new_value
        return updated_component



from copy import deepcopy
from importlib.resources import files
from typing import Any, Dict, List

from .. import RELAIData
from ...utils import format_title


class BarPlot(RELAIData):
    """Data model for a bar plot component.

    Args:
        title (str): The main title of the chart.
        x_axis_label (str): The label for the chart's X-axis.
        y_axis_label (str): The label for the chart's Y-axis.
        data_key (str): The key in each data object that represents the primary label for the bar.
        data (List[Dict[str, Any]]): An array of data objects to be plotted. Each object must contain at least
            the data_key and one numeric value.
        column_span (int, optional): Number of columns the component should span. Defaults to 12.

    Examples:
        >>> # Create a simple bar plot showing performance by benchmark
        >>> data = [
        >>>     { "benchmark": "Summarization", "score": 0.6203 },
                >>>     { "benchmark": "Q&A Standard", "score": 0.5452 },
                >>>     { "benchmark": "Q&A Reasoning", "score": 0.7738 },
        >>> ]
        >>> bar_plot = BarPlot(
        >>>     title="Performance comparison across all benchmarks",
        >>>     x_axis_label="Benchmark",
        >>>     y_axis_label="Score",
        >>>     data_key="benchmark",
        >>>     data=data
        >>> )
    """

    schema_file: str = files("relai_visualization").joinpath("schema/components/bar-plot.schema.json")

    def __init__(
        self,
        title: str,
        x_axis_label: str,
        y_axis_label: str,
        data_key: str,
        data: List[Dict[str, Any]],
        column_span: int = 12,
    ) -> None:
        # Make a deep copy of the data to avoid modifying the original
        data = deepcopy(data)

        super().__init__(
            componentName="plot",
            type="bar",
            title=format_title(title),
            xAxisLabel=x_axis_label,
            yAxisLabel=y_axis_label,
            dataKey=data_key,
            data=data,
            columnSpan=column_span,
        )

    def custom_validation(self) -> None:
        """Custom validation for the bar plot data.

        Validates that:
        1. Data is not empty
        2. All data points contain the specified data_key
        3. Each data point has at least one numeric value
        """
        # Check that data is not empty
        if not self.data["data"]:
            raise ValueError("Data cannot be empty")

        data_key = self.data["dataKey"]

        for item in self.data["data"]:
            # Check that the item contains the data_key
            if data_key not in item:
                raise ValueError(f"All data points must contain the key '{data_key}'")

            # Ensure each data point has at least one other property (numeric value)
            if len(item) < 2:
                raise ValueError("Each data point must have at least one numeric value in addition to the data key")

            # Check that there's at least one numeric value
            has_numeric = False
            for key, value in item.items():
                if key != data_key and isinstance(value, (int, float)):
                    has_numeric = True
                    break

            if not has_numeric:
                raise ValueError(f"Data point '{item}' does not contain any numeric values")



import re
from copy import deepcopy
from importlib import import_module
from importlib.resources import files
from typing import Any

from .relai_data import RELAIData


def get_component_cls_by_name(component_name: str) -> RELAIData:
    component_cls_name = None
    if component_name == "feedback-question":
        component_cls_name = "FeedbackQuestion"
    elif component_name == "feedback-questions":
        component_cls_name = "FeedbackQuestions"
    elif component_name == "image-grid":
        component_cls_name = "ImageGrid"
    elif component_name == "image":
        component_cls_name = "Image"
    elif component_name == "table":
        component_cls_name = "Table"
    elif component_name == "vega-chart":
        component_cls_name = "VegaChart"
    else:
        raise ValueError(f"Unrecognized component with name '{component_name}'.")
    return getattr(import_module(".components", package=__package__), component_cls_name)


class ApplicationResult(RELAIData):
    """Data model for an application result.
    See the schema at https://api.relai.ai/schema/application-result.schema.json for details about the accepted values.

    Args:
        application (str): Name of the application.
        application_version (str): Version of the application.
        display_name (str): Display name of the application.
        application_info (str): A short description of the application.
        started_at (str): Timestamp when the run was started.
        project_mame (str): Name of the project.
        run_id (str): ID of the run.
        components (dict[str, RELAIData]): Dictionary mapping component ids to corresponding components.
        generated_by (str, optional): Name of the user that generated the application result. Defaults to None.
        ended_at (str, optional): Timestamp when the run was ended. Defaults to None.
        schema_version (str, optional): Schema version. Defaults to "v1.0".
        metadata (dict, optional): Additional metadata. Defaults to None.
    """

    schema_file: str = files("relai_visualization").joinpath("schema/application-result.schema.json")

    def __init__(
        self,
        application: str,
        application_version: str,
        display_name: str,
        application_info: str,
        started_at: str,
        project_name: str,
        run_id: str,
        components: dict[str, RELAIData],
        generated_by: str = None,
        ended_at: str = None,
        schema_version: str = "v1.0",
        metadata: dict = None,
    ) -> None:
        if not re.match(r"^[a-z0-9-]+$", application):
            raise ValueError("Application name must only include lowercase letters, digits and hyphens.")
        if not re.match(r"^v\d+\.\d+$", application_version):
            raise ValueError("Application version must be in the format 'v{x}.{y}'.")

        if not re.match(r"^[a-z0-9-_]+$", project_name):
            raise ValueError("Project name must only include lowercase characters, hyphens, and underscores.")

        if len(components) < 1:
            raise ValueError("At least one component must be provided.")

        for id, component in components.items():
            components[id] = component.data

        super().__init__(
            schemaVersion=schema_version,
            application=application,
            applicationVersion=application_version,
            displayName=display_name,
            applicationInfo=application_info,
            generatedBy=generated_by,
            startedAt=started_at,
            endedAt=ended_at,
            projectName=project_name,
            runId=run_id,
            metadata=metadata,
            components=components,
        )

    @classmethod
    def user_feedback(cls, component: dict) -> dict:
        """Get the feedback provided by the user for the component.

        Args:
            component (dict): ApplicationResult component data as a dictionary.

        Returns:
            dict: User feedback for the component as a dictionary mapping component id to feedback values.
                Feedback for components that do not support feedback will be `None`.
        """
        output = {}
        for component_id, component_data in component["components"].items():
            try:
                output[component_id] = get_component_cls_by_name(component_data["componentName"]).user_feedback(
                    component_data
                )
            except NotImplementedError:
                output[component_id] = None
        return output

    @classmethod
    def update_with_feedback(cls, component: dict, modifies: list[str], new_value: Any) -> dict:
        """Update an ApplicationResult with user feedback.

        Args:
            component (dict): ApplicationResult component to update as a Python dictionary.
            modifies (list[str]): List of strings that uniquely identify the data to be updated.
                The first element must be the component id. The rest of the element must be identifiers for within
                the component.
            new_value (Any): New value to set.

        Returns:
            dict: Updated ApplicationResult.
        """
        component = deepcopy(component)
        component_id = modifies[0]
        try:
            component_to_update = component["components"][component_id]
        except KeyError:
            raise ValueError(f"Component with id '{component_id}' not found.")
        component_name = component_to_update["componentName"]
        try:
            component["components"][component_id] = get_component_cls_by_name(component_name).update_with_feedback(
                component_to_update, modifies[1:], new_value
            )
        except NotImplementedError:
            raise RuntimeError(
                f"Component with id '{component_id}' and name '{component_name}' does not support feedback."
            )
        return component



from .models import ApplicationResult, RELAIData
from .models.components import BarPlot, FeedbackQuestion, FeedbackQuestions, Image, ImageGrid, Table, VegaChart
from .models.data import Array2D, DataType

__all__ = [
    "ApplicationResult",
    "Array2D",
    "BarPlot",
    "DataType",
    "FeedbackQuestion",
    "FeedbackQuestions",
    "Image",
    "ImageGrid",
    "RELAIData",
    "Table",
    "VegaChart",
]



def format_title(column_name: str) -> str:
    """Makes titles more readable.

    Converts snake_case to Title Case with spaces.

    Args:
        column_name (str): The column name in snake_case.

    Returns:
        str: The formatted column name in Title Case.
    """
    column_name = column_name.replace("_", " ").replace("-", " ")
    return " ".join(word.capitalize() for word in column_name.split())



import base64
import io
import json
from copy import deepcopy
from importlib.resources import files
from urllib.parse import urlparse

from PIL import Image as PILImage

from .. import RELAIData


def _get_relai_client_cls():
    try:
        from relai_core import RELAIClient
    except ModuleNotFoundError as exc:
        raise ModuleNotFoundError(
            "Image uploads via `upload_to_run_id` require a compatible `relai-core` installation."
        ) from exc
    return RELAIClient


def is_valid_url(url: str) -> bool:
    """Check if the URL is valid.

    Args:
        url (str): URL to be validated.

    Returns:
        bool: True if the URL is valid, False otherwise.
    """
    parsed_url = urlparse(url)
    return all([parsed_url.scheme, parsed_url.netloc])


class Image(RELAIData):
    """Data model for an image component. Either `image_url` or `image` must be provided.

    Args:
        image_url (str, optional): URL of the image. Defaults to None.
        image (PILImage.Image, optional): Image object. Defaults to None.
        upload_to_run_id (str, optional): Run ID to which the image should be uploaded.
            Setting this parameter will first upload the image to the specified run ID and then use the image URL.
            This significantly reduces the size of the JSON payload. Defaults to None.
        caption (str, optional): Caption for the image. Defaults to None.
        column_span (int, optional): Number of columns the component should span. Defaults to 12.
    """

    schema_file: str = files("relai_visualization").joinpath("schema/components/image.schema.json")

    def __init__(
        self,
        image_url: str = None,
        image: PILImage.Image = None,
        upload_to_run_id: str = None,
        caption: str = None,
        column_span: int = 12,
    ) -> None:
        if (image_url and image) or (not image_url and not image):
            raise ValueError("Exactly one of `image_url` or `image` must be provided.")

        if image_url and upload_to_run_id:
            raise ValueError("Cannot provide both `image_url` and `upload_to_run_id`.")

        image_data = None
        if image:
            # Upload image and retrieve URL
            if upload_to_run_id:
                relai_client = _get_relai_client_cls()()
                image_url = relai_client.upload_image(image, run_id=upload_to_run_id)
            else:
                buffer = io.BytesIO()
                image.save(buffer, format="png")
                image_data = base64.b64encode(buffer.getvalue()).decode("utf-8")

        super().__init__(
            componentName="image",
            columnSpan=column_span,
            src=image_url,
            data=image_data,
            caption=caption,
        )

    def custom_validation(self) -> None:
        try:
            if not is_valid_url(self.data["src"]):
                raise ValueError("Invalid URL")
        except KeyError:
            # Skip validation if src is not provided
            pass

    def __str__(self) -> str:
        output_dict = deepcopy(self.data)
        try:
            output_dict["data"] = f"Base64 encoded image: '{output_dict['data'][:10]}...'"
        except KeyError:
            pass
        return json.dumps(output_dict)



from .bar_plot import BarPlot
from .feedback_question import FeedbackQuestion
from .feedback_questions import FeedbackQuestions
from .image import Image
from .image_grid import ImageGrid
from .table import Table
from .vega_chart import VegaChart

__all__ = [
    "BarPlot",
    "FeedbackQuestion",
    "FeedbackQuestions",
    "Image",
    "ImageGrid",
    "Table",
    "VegaChart",
]



from importlib.resources import files

import numpy as np

from .. import RELAIData


class Array2D(RELAIData):
    """Data model for a 2D array.

    Args:
        array (np.ndarray): 2D array to be stored.
    """

    schema_file: str = files("relai_visualization").joinpath("schema/data/2D-array.schema.json")

    def __init__(self, array: np.ndarray) -> None:
        if array.ndim != 2:
            raise ValueError("Array must be 2D.")
        if not np.issubdtype(array.dtype, np.number):
            raise ValueError("Array must be numeric.")
        if array.size == 0:
            raise ValueError("Array must not be empty.")
        super().__init__(data=array.tolist())



import json
import os
from pathlib import Path
from typing import Any, Union

import requests
from jsonschema import validate
from jsonschema.exceptions import ValidationError
from referencing import Registry, Resource


def fetch_remote(uri):
    response = requests.get(uri)
    response.raise_for_status()
    return Resource.from_contents(response.json())


class RELAIData:
    """Base class for RELAI data models. Validations that are performed directly on data passed to a child class
    should be done before calling `__init__` of this base class.

    To enable strict schema validation, set the environment variable `RELAI_DEVELOPMENT_MODE` to `DEBUG`.

    Args:
        **kwargs: Data to be validated against the schema.
    """

    schema_file: Union[str, Path] = None  # Path to the schema file for this component

    def __init__(self, **kwargs) -> None:
        with open(self.schema_file, "r") as f:
            self._schema = json.load(f)

        self._data = kwargs
        # Remove keys with None values
        self._data = {k: v for k, v in self._data.items() if v is not None}

        # Remote schema registry
        self._remote_schema_registry = Registry(
            resources={
                "vega": "https://vega.github.io/schema/vega/v6.json",
                "vega-lite": "https://vega.github.io/schema/vega-lite/v6.json",
            },
            retrieve=fetch_remote,
        )

        relai_development_mode = os.getenv("RELAI_DEVELOPMENT_MODE", "DEBUG")
        if relai_development_mode == "DEBUG":
            self.schema_validation()
            self.custom_validation()
        elif relai_development_mode != "PRODUCTION":
            raise ValueError("Invalid value for RELAI_DEVELOPMENT_MODE. Must be 'DEBUG' or 'PRODUCTION'.")

    @property
    def data(self) -> dict:
        """Get the data. This property is read-only."""
        return self._data

    def schema_validation(self) -> None:
        """Validate the data against the schema."""
        try:
            validate(
                instance=self._data,
                schema=self._schema,
                registry=self._remote_schema_registry,
            )
        except ValidationError as e:
            raise ValueError(f"Data validation failed: {e.message}") from e

    def custom_validation(self) -> None:
        """Custom validation logic. Use this to validate the JSON data dict against custom rules,
        after schema validation."""
        pass

    def json(self) -> str:
        """Serialize the data to a JSON formatted `str`."""
        return json.dumps(self._data)

    @classmethod
    def user_feedback(cls, component: dict) -> dict:
        """Get the feedback provided by the user for the component.

        Args:
            component (dict): Component data.
        """
        raise NotImplementedError

    def __str__(self) -> str:
        return self.json()

    @classmethod
    def update_with_feedback(cls, component: dict, modifies: list[str], new_value: Any) -> dict:
        """Update a `RELAIData` component with user feedback.

        Args:
            component (dict): Component to update as a Python dictionary.
            modifies (list[str]): List of strings that uniquely identify the data to be updated.
            new_value (Any): New value to set.

        Returns:
            dict: Updated component.
        """
        raise NotImplementedError

    @classmethod
    def from_data(cls, data: dict) -> "RELAIData":
        """Create a new instance of the class from the data dictionary.

        Args:
            data (dict): Data dictionary.

        Returns:
            RELAIData: Instance of the class.
        """
        component = cls.__new__(cls)
        super(cls, component).__init__(**data)
        return component



import json
from copy import deepcopy
from importlib.resources import files

from .. import RELAIData
from .image import Image as RELAIImage


class ImageGrid(RELAIData):
    """Data model for an image grid component.

    Args:
        images (list[Image]): List of RELAI images to be displayed in the grid.
        grid_size (tuple[int, int]): Size of the grid in terms of rows and columns.
        row_captions (list[str], optional): List of row captions. Defaults to None.
        column_captions (list[str], optional): List of column captions. Defaults to None.
        caption (str, optional): Caption for the image grid. Defaults to None.
        columnSpan (int, optional): Number of columns the component should span. Defaults to 12.
    """

    schema_file: str = files("relai_visualization").joinpath("schema/components/image-grid.schema.json")

    def __init__(
        self,
        images: list[RELAIImage],
        grid_size: tuple[int, int],
        row_captions: list[str] = None,
        column_captions: list[str] = None,
        caption: str = None,
        column_span: int = 12,
    ) -> None:
        if len(grid_size) != 2 or not all(isinstance(i, int) for i in grid_size):
            raise ValueError("Grid size must be a tuple of two integers")

        if grid_size[0] < 1 or grid_size[1] < 1:
            raise ValueError("Grid parameters must be greater than 0")

        if row_captions and len(row_captions) < 1:
            raise ValueError("Row captions cannot be empty when provided.")

        if column_captions and len(column_captions) < 1:
            raise ValueError("Column captions cannot be empty when provided.")

        self._image_objects = images
        images = [image.data for image in self._image_objects]
        grid_size = list(grid_size)

        super().__init__(
            componentName="image-grid",
            columnSpan=column_span,
            images=images,
            grid_size=grid_size,
            row_captions=row_captions,
            column_captions=column_captions,
            caption=caption,
        )

    def custom_validation(self) -> None:
        num_rows, num_cols = self.data["grid_size"]
        num_images = len(self.data["images"])
        if not (
            num_images == num_cols * num_rows  # Exact number of images
            or (num_images // num_cols == num_rows - 1 and num_images % num_cols != 0)  # Last row incomplete
        ):
            raise ValueError("Grid is too small or too large for the number of images provided.")
        try:
            if len(self.data["row_captions"]) != num_rows:
                raise ValueError("Number of row captions must match the number of rows in the grid.")
        except KeyError:
            # Skip validation if row captions are not provided
            pass
        try:
            if len(self.data["column_captions"]) != num_cols:
                raise ValueError("Number of column captions must match the number of columns in the grid.")
        except KeyError:
            # Skip validation if column captions are not provided
            pass

    def __str__(self) -> str:
        output_dict = deepcopy(self.data)
        output_dict["images"] = [json.loads(str(image)) for image in self._image_objects]
        return json.dumps(output_dict)



import uuid
from copy import deepcopy
from importlib.resources import files
from typing import Any

from .. import RELAIData
from ..data import DataType
from ...utils import format_title
from .feedback_questions import FeedbackQuestions


def format_row_value(value: Any) -> str:
    """Makes the row value more readable.

    Converts lists and dictionaries to a string representation.

    Args:
        value (Any): The value to format.

    Returns:
        str: The formatted value.
    """
    if isinstance(value, list):
        return "\n".join(map(str, value))
    elif isinstance(value, dict):
        return "\n".join(f"{k}: {v}" for k, v in value.items())
    return value


class Table(RELAIData):
    """Data model for a table component.

    Args:
        title (str): Title of the table.
        columns (list[Any]): List of columns. Each column must follow the following format:
            {
                "name": str,
                "type": `DataType` object,
                "hideInTable": bool (optional, default: False),
                "showInModal": bool (optional, default: True),
            }
        rows (list[Any]): List of rows. Each row is a dictionary with keys corresponding to column names.
        accepts_feedback (bool, optional): Whether this component accepts feedback. Defaults to False.
        feedback_questions (list[Any], optional): List of feedback questions. Defaults to None.
            Each feedback question must follow the following format:
                {
                    "question": str,
                    "type": `DataType` object,
                    "context": A list of RELAIData components to show as context for the question. Required if the
                        type is "contextSelection",
                    "forRow": int (0-based row index for which the question is asked),
                    "forColumn": int (0-based column index for which the question is asked). Can be `None`
                        if the question is not specific to a column,
                }
        column_span (int, optional): Number of columns the component should span. Defaults to 12.
    """

    schema_file: str = files("relai_visualization").joinpath("schema/components/table.schema.json")

    def __init__(
        self,
        title: str,
        columns: list[Any],
        rows: list[Any],
        accepts_feedback: bool = False,
        feedback_questions: list[Any] = None,
        column_span: int = 12,
    ) -> None:
        if accepts_feedback:
            if not feedback_questions:
                raise ValueError("Feedback questions must be provided if feedback is accepted.")
            if len(feedback_questions) < 1:
                raise ValueError("At least one feedback question must be provided.")

        formatted_rows = []
        columns = deepcopy(columns)

        # Add unique IDs to columns and format their names
        for column in columns:
            if "rowId" in column:
                raise ValueError("'rowId' field is reserved and not allowed.")
            column["columnId"] = f"column-{uuid.uuid4().hex}"
            column["name"] = format_title(column["name"])

        # Add unique IDs to rows and format their values
        for row in rows:
            new_row = {}
            for k, v in row.items():
                new_row[format_title(k)] = format_row_value(v)
            new_row["rowId"] = f"row-{uuid.uuid4().hex}"
            formatted_rows.append(new_row)
        rows = formatted_rows

        rowwise_feedback_questions = None
        if feedback_questions:
            for feedback_question in feedback_questions:
                try:
                    if columns[feedback_question["forColumn"]]["type"] != feedback_question["type"]:
                        raise ValueError("Feedback question type must match the type of the column it modifies.")
                except IndexError:
                    raise ValueError("Invalid 'forColumn' index in feedback question.")
                except TypeError:
                    # If forColumn is None, the question is not specific to a column
                    pass
            rowwise_feedback_questions = [[] for _ in range(len(rows))]
            for feedback_question in feedback_questions:
                try:
                    if not 0 <= feedback_question["forRow"] < len(rows):
                        raise ValueError("Invalid 'forRow' index in feedback question.")
                except KeyError:
                    raise ValueError("Feedback question must have a 'forRow' field.")
                rowwise_feedback_questions[feedback_question["forRow"]].append(
                    {
                        "question": feedback_question["question"],
                        "type": feedback_question["type"],
                        "context": feedback_question.get("context"),
                        "modifies": [
                            rows[feedback_question["forRow"]]["rowId"],
                        ],
                    }
                )
                # If the feedback question is specific to a column, add the column ID to the modifies list
                if feedback_question["forColumn"] is not None:
                    rowwise_feedback_questions[feedback_question["forRow"]][-1]["modifies"].append(
                        columns[feedback_question["forColumn"]]["columnId"]
                    )
            rowwise_feedback_questions = [
                FeedbackQuestions(row_questions).data for row_questions in rowwise_feedback_questions
            ]

        for column in columns:
            column["type"] = column["type"].data

        super().__init__(
            componentName="table",
            title=title,
            columnSpan=column_span,
            columns=columns,
            rows=rows,
            acceptsFeedback=accepts_feedback,
            feedbackQuestions=rowwise_feedback_questions,
        )

    def custom_validation(self) -> None:
        for row in self.data["rows"]:
            if set(row.keys()) - {"rowId"} != set(c["name"] for c in self.data["columns"]):
                raise ValueError("Column names and row fields must match.")
            for column in self.data["columns"]:
                for row in self.data["rows"]:
                    value = row[column["name"]]
                    # Check if the value is of valid type, as long as it is not None
                    if value and not DataType.from_dict(column["type"]).validate(value):
                        raise ValueError(f"Invalid value '{value}' for column '{column['name']}'")

    @classmethod
    def user_feedback(cls, component: dict) -> list[dict]:
        """Get the user feedback for the feedback questions.

        Args:
            component (dict): Table component data as a dictionary.

        Returns:
            dict: User feedback for the feedback questions as a list of dictionaries mapping question to feedback values.
                Each item in the list corresponds to a row in the table in the same order.
        """
        output = [{} for _ in range(len(component["feedbackQuestions"]))]
        for idx, row_questions in enumerate(component["feedbackQuestions"]):
            for question_id, answer in row_questions["feedback"].items():
                output[idx][row_questions["feedbackQuestions"][question_id]["question"]] = answer
        return output

    @classmethod
    def update_with_feedback(cls, component: dict, modifies: list[str], new_value: Any) -> dict:
        """Update a `Table` component with user feedback.

        Args:
            component (dict): `Table` component to update as a Python dictionary.
            modifies (list[str]): QuestionId, RowId and ColumnId to modify.
            new_value (Any): New value to set.

        Returns:
            dict: Updated `Table` component.
        """
        component = deepcopy(component)
        try:
            question_id, row_id, column_id = modifies
        except ValueError:
            # column_id is missing
            question_id, row_id = modifies
            column_id = None

        # Find the row index and update the value
        for idx, row in enumerate(component["rows"]):
            if row["rowId"] == row_id:
                row_idx = idx
                break
        else:
            raise ValueError("Invalid row ID.")

        if column_id is not None:
            # Do this only if the feedback is for a specific column
            # Find the column name and index
            for idx, column in enumerate(component["columns"]):
                if column["columnId"] == column_id:
                    column_name = column["name"]
                    column_idx = idx
                    break
            else:
                raise ValueError("Invalid column ID.")

            # Check if the value is of valid type, as long as it is not None
            if new_value and not DataType.from_dict(component["columns"][column_idx]["type"]).validate(new_value):
                raise ValueError(f"Invalid value '{new_value}' for column {column_idx}.")

            # Update the value
            component["rows"][row_idx][column_name] = new_value

        # Update the FeedbackQuestions component with the new feedback value as well
        component["feedbackQuestions"][row_idx] = FeedbackQuestions.update_with_feedback(
            component=component["feedbackQuestions"][row_idx],
            modifies=modifies,
            new_value=new_value,
        )
        return component



from importlib.resources import files

from .. import RELAIData


class VegaChart(RELAIData):
    """Data model for a Vega chart component.

    Args:
        spec (dict): Vega chart specification.
        column_span (int, optional): Number of columns the component should span. Defaults to 12.
    """

    schema_file: str = files("relai_visualization").joinpath("schema/components/vega-chart.schema.json")

    def __init__(self, spec: dict, column_span: int = 12, vega_lite=True) -> None:
        super().__init__(
            componentName="vega-chart",
            columnSpan=column_span,
            vegaLite=vega_lite,
            spec=spec,
        )

    def custom_validation(self) -> None:
        if self.data["vegaLite"] and not self.data["spec"]["$schema"].startswith(
            "https://vega.github.io/schema/vega-lite/"
        ):
            raise ValueError("Invalid Vega-Lite specification.")
        if not self.data["vegaLite"] and not self.data["spec"]["$schema"].startswith(
            "https://vega.github.io/schema/vega/"
        ):
            raise ValueError("Invalid Vega specification.")



import json
import warnings
from copy import deepcopy
from importlib.resources import files
from typing import Any

from .. import RELAIData


class DataType(RELAIData):
    """Data model for a data type.

    Args:
        data_type (str): Type of the data. Must be one of "string", "integer", "float", "boolean", or "enum".
        options (list[str], optional): List of options for the data type. Defaults to None.
            Must be provided if `data_type` is "enum".
    """

    schema_file: str = files("relai_visualization").joinpath("schema/data/data-type.schema.json")

    def __init__(self, data_type: str, options: list[str] = None) -> None:
        # TODO: Drop support for "number" data type
        if data_type not in ["string", "integer", "float", "boolean", "enum", "number"]:
            raise ValueError("Invalid data type")
        if data_type == "number":
            warnings.warn(
                """'number' data type is deprecated and will be removed in a future version."""
                """Use 'integer' or 'float' instead. For now, 'number' is treated as 'float'.""",
                DeprecationWarning,
                stacklevel=2,
            )
            data_type = "float"
        if (data_type == "enum" and not options) or (options and data_type != "enum"):
            raise ValueError("Options must and only be provided for `enum` data type")
        if data_type == "enum" and len(options) < 2:
            raise ValueError("At least two options must be provided for `enum` data type")
        super().__init__(type=data_type, options=options)

    def __str__(self) -> str:
        output_dict = deepcopy(self.data)
        if self.data["type"] == "enum":
            output_dict["type"] = f"enum({self.data['options']})"
            del output_dict["options"]
        return json.dumps(output_dict)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DataType):
            return False
        return self.data == other.data

    def validate(self, value: Any) -> bool:
        """Check if the value is valid for the data type.

        Args:
            value (Any): Value to be validated.

        Returns:
            bool: True if the value is valid, False otherwise.
        """
        if self.data["type"] == "string":
            return isinstance(value, str)
        if self.data["type"] == "integer":
            return isinstance(value, int)
        if self.data["type"] == "float":
            return isinstance(value, float)
        if self.data["type"] == "boolean":
            return isinstance(value, bool)
        if self.data["type"] == "enum":
            return value in self.data["options"]
        return False

    @classmethod
    def from_dict(cls, data: dict) -> "DataType":
        """Create a data type from a dictionary."""
        return cls(data["type"], data.get("options"))



from importlib.resources import files
from typing import Union

from .. import RELAIData
from ..data.data_type import DataType


class FeedbackQuestion(RELAIData):
    """Data model for a feedback question component.

    Args:
        question (str): Question to be asked.
        type (Union[DataType, str]): Type of the question. Must be either a `DataType` object
            or the string "contextSelection".
        modifies (list[str]): A list of one or more strings that uniquely determine what needs to be changed in the application result.
        context (list[RELAIData], optional): A list of RELAIData components to show as context for the question.
            Defaults to None.
        column_span (int, optional): Number of columns the component should span. Defaults to 12.
    """

    schema_file: str = files("relai_visualization").joinpath("schema/components/feedback-question.schema.json")

    def __init__(
        self,
        question: str,
        type: Union[DataType, str],
        modifies: list[str],
        context: list[RELAIData] = None,
        column_span: int = 12,
    ) -> None:
        if context:
            # TODO: This check is broader than what is allowed by the schema.
            # They should be matched once the schema is updated.
            if len(context) < 1:
                raise ValueError("Context cannot be empty when provided.")
            if any([not isinstance(c, RELAIData) for c in context]):
                raise ValueError("Invalid context. Make sure all elements in `context` are of type RELAIData.")
            context = [c.data for c in context]

        if isinstance(type, DataType):
            type = type.data
        elif type != "contextSelection":
            raise ValueError("Invalid type.")

        super().__init__(
            componentName="feedback-question",
            columnSpan=column_span,
            question=question,
            context=context,
            type=type,
            modifies=modifies,
        )
